#!/usr/bin/python

# This script requires i3ipc-python package (install it from a system package manager
# or pip).
# It adds icons to the workspace name for each open window.
# Set your keybindings like this: set $workspace1 workspace number 1
# Add your icons to WINDOW_ICONS.
# Based on https://github.com/maximbaz/dotfiles/blob/master/bin/i3-autoname-workspaces

import argparse
import logging
import re
import signal
import sys

import i3ipc

WINDOW_ICONS = {
    "firefox": "",
}

DEFAULT_ICON = "󰀏"


def icon_for_window(window):
    name = None
    if window.app_id is not None and len(window.app_id) > 0:
        name = window.app_id.lower()
    elif window.window_class is not None and len(window.window_class) > 0:
        name =  window.window_class.lower()

    if name in WINDOW_ICONS:
        return WINDOW_ICONS[name]

    logging.info("No icon available for window with name: %s" % str(name))
    return DEFAULT_ICON

def rename_workspaces(ipc):
    for workspace in ipc.get_tree().workspaces():
        name_parts = parse_workspace_name(workspace.name)
        icon_tuple = ()
        for w in workspace:
            if w.app_id is not None or w.window_class is not None:
                icon = icon_for_window(w)
                if not ARGUMENTS.duplicates and icon in icon_tuple:
                    continue
                icon_tuple += (icon,)
        name_parts["icons"] = "  ".join(icon_tuple) + " "
        new_name = construct_workspace_name(name_parts)
        ipc.command('rename workspace "%s" to "%s"' % (workspace.name, new_name))


def undo_window_renaming(ipc):
    for workspace in ipc.get_tree().workspaces():
        name_parts = parse_workspace_name(workspace.name)
        name_parts["icons"] = None
        new_name = construct_workspace_name(name_parts)
        ipc.command('rename workspace "%s" to "%s"' % (workspace.name, new_name))
    ipc.main_quit()
    sys.exit(0)


def parse_workspace_name(name):
    return re.match(
        r"(?P<num>[0-9]+):?(?P<shortname>\w+)? ?(?P<icons>.+)?", name
    ).groupdict()


def construct_workspace_name(parts):
    new_name = str(parts["num"])
    if parts["shortname"] or parts["icons"]:
        new_name += ":"

        if parts["shortname"]:
            new_name += parts["shortname"]

        if parts["icons"]:
            new_name += " " + parts["icons"]

    return new_name


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="This script automatically changes the workspace name in sway depending on your open applications."
    )
    parser.add_argument(
        "--duplicates",
        "-d",
        action="store_true",
        help="Set it when you want an icon for each instance of the same application per workspace.",
    )
    parser.add_argument(
        "--logfile",
        "-l",
        type=str,
        default="/tmp/sway-autoname-workspaces.log",
        help="Path for the logfile.",
    )
    args = parser.parse_args()
    global ARGUMENTS
    ARGUMENTS = args

    logging.basicConfig(
        level=logging.INFO,
        filename=ARGUMENTS.logfile,
        filemode="w",
        format="%(message)s",
    )

    ipc = i3ipc.Connection()

    for sig in [signal.SIGINT, signal.SIGTERM]:
        signal.signal(sig, lambda signal, frame: undo_window_renaming(ipc))

    def window_event_handler(ipc, e):
        if e.change in ["new", "close", "move"]:
            rename_workspaces(ipc)

    ipc.on("window", window_event_handler)

    rename_workspaces(ipc)

    ipc.main()




#!/usr/bin/python

# This script requires i3ipc-python package (install it from a system package manager
# or pip).
# It makes inactive windows transparent. Use `transparency_val` variable to control
# transparency strength in range of 0…1 or use the command line argument -o.

import argparse
import signal
import sys
from functools import partial

import i3ipc


def on_window(args, ipc, event):
    global focused_set

    # To get the workspace for a container, we need to have received its
    # parents, so fetch the whole tree
    tree = ipc.get_tree()

    focused = tree.find_focused()
    if focused is None:
        return

    focused_workspace = focused.workspace()

    focused.command("opacity " + args.focused)
    focused_set.add(focused.id)

    to_remove = set()
    for window_id in focused_set:
        if window_id == focused.id:
            continue
        window = tree.find_by_id(window_id)
        if window is None:
            to_remove.add(window_id)
        elif args.global_focus or window.workspace() == focused_workspace:
            window.command("opacity " + args.opacity)
            to_remove.add(window_id)

    focused_set -= to_remove

def remove_opacity(ipc, focused_opacity):
    for workspace in ipc.get_tree().workspaces():
        for w in workspace:
            w.command("opacity " + focused_opacity)
    ipc.main_quit()
    sys.exit(0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="This script allows you to set the transparency of unfocused windows in sway."
    )
    parser.add_argument(
        "--opacity",
        "-o",
        type=str,
        default="0.80",
        help="set inactive opacity value in range 0...1",
    )
    parser.add_argument(
        "--focused",
        "-f",
        type=str,
        default="1.0",
        help="set focused opacity value in range 0...1",
    )
    parser.add_argument(
        "--global-focus",
        "-g",
        action="store_true",
        help="only have one opaque window across all workspaces",
    )
    args = parser.parse_args()

    ipc = i3ipc.Connection()
    focused_set = set()

    for window in ipc.get_tree():
        if window.focused:
            focused_set.add(window.id)
            window.command("opacity " + args.focused)
        else:
            window.command("opacity " + args.opacity)
    for sig in [signal.SIGINT, signal.SIGTERM]:
        signal.signal(sig, lambda signal, frame: remove_opacity(ipc, args.focused))
    ipc.on("window", partial(on_window, args))
    ipc.main()



"""
Utility to selectively disable keypresses to specific windows.

This program was written due to Firefox's pop-out video player closing when
the Escape key is pressed.  I use a modal text editor (Helix) that encourages
regularly pressing that key and it had a habit of going to the wrong window.

The easiest way I could find to make this window specific key-binding change was
via this code.  Specifically it watches focus changes until the "right" windowds
is focused, and then causes Sway to bind the Escape key before Firefox can see
it.  It continues to watch focus changes so that this binding can be disabled
when another window is selected.

This feels like a potentially useful pattern, please let us know:
  https://github.com/OctopusET/sway-contrib
of any other programs that this functionality would benefit.
"""

import argparse
import logging
import signal
from dataclasses import dataclass
from typing import Any

import i3ipc

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class Watch:
    container_props: dict[str,str]
    binds: set[str]


class Monitor:
    _bound: set[str]
    watched: list[Watch]

    def __init__(self) -> None:
        self.ipc = i3ipc.Connection()
        self.ipc.on("window::focus", self.on_window_event)
        # firefox creates PIP window without title, so need to watch for changes
        self.ipc.on("window::title", self.on_window_event)
        self._bound = set()
        self.watched = []

    def bind(self, *binds: str, **props: str) -> None:
        self.watched.append(Watch(props, set(binds)))

    def run(self) -> None:
        "run main i3ipc event loop"
        ipc = self.ipc
        def sighandler(signum: int, frame: Any) -> None:
            logger.debug("exit signal received, stopping event loop")
            ipc.main_quit()

        # stop event loop when we get one of these
        for sig in signal.SIGINT, signal.SIGTERM:
            signal.signal(sig, sighandler)

        try:
            ipc.main()
        finally:
            # clean up
            self.bound = set()

    def on_window_event(self, ipc: i3ipc.Connection, event: i3ipc.WindowEvent) -> None:
        "respond to window events"
        container = event.container
        if not container.focused:
            return
        data = container.ipc_data
        logger.debug("window event %s", data)
        binds = set()
        for watch in self.watched:
            if all(data.get(k) == v for k, v in watch.container_props.items()):
                    binds.update(watch.binds)

        self.bound = binds

    @property
    def bound(self) -> set[str]:
        return self._bound

    @bound.setter
    def bound(self, binds: set[str]) -> None:
        if binds == self._bound:
            return
        to_del = self._bound - binds
        to_add = binds - self._bound
        if to_del:
            logger.info(f"removing binds {', '.join(to_del)}")
        if to_add:
            logger.info(f"adding binds {', '.join(to_add)}")
        for bind in to_del:
            self.ipc.command(f"unbindsym {bind}")
        for bind in to_add:
            msg = f"{bind} ignored due to focus monitor"
            self.ipc.command(f"bindsym {bind} exec echo '{msg}'")
        self._bound = binds


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="track active window to disable Esc key in Firefox popout media player",
    )
    parser.add_argument(
        "--verbose", "-v", help="Increase verbosity", action="store_true"
    )
    args = parser.parse_args()
    args.loglevel = logging.DEBUG if args.verbose else logging.INFO
    return args


KEY_ESCAPE = "Escape"


def main() -> None:
    args = parse_args()
    logging.basicConfig(
        level=args.loglevel,
        format="%(asctime)s %(levelname)s %(message)s",
    )

    mon = Monitor()
    # block Escape key from reaching Firefox's popout window
    mon.bind(KEY_ESCAPE, app_id="firefox", name="Picture-in-Picture")
    mon.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# This script keeps track of the active layout for each window.
# Optional argument defines numeric layout index for new windows (counted from 0)
#
# This script requires i3ipc-python package (install it from a system package
# manager or pip).

import sys
from typing import Optional

import i3ipc


def on_window_focus(ipc: i3ipc.connection.Connection, event: i3ipc.events.WindowEvent):
    global windows, prev_focused, default_layout

    # Get current layouts
    layouts = {
        input.identifier: input.xkb_active_layout_index for input in ipc.get_inputs()
    }

    # Save layouts for previous window
    windows[prev_focused] = layouts

    # Restore layout of the newly focused known window
    if event.container.id in windows:
        for kdb_id, layout_index in windows[event.container.id].items():
            if layout_index != layouts[kdb_id]:
                ipc.command(f'input "{kdb_id}" xkb_switch_layout {layout_index}')
                break

    # Set default layout for a fresh window
    elif default_layout is not None:
        for kdb_id, layout_index in layouts.items():
            if layout_index is not None and layout_index != default_layout:
                ipc.command(f'input "{kdb_id}" xkb_switch_layout {default_layout}')
                break

    prev_focused = event.container.id


def on_window_close(ipc: i3ipc.connection.Connection, event: i3ipc.events.WindowEvent):
    global windows
    if event.container.id in windows:
        del windows[event.container.id]


def on_window(ipc: i3ipc.connection.Connection, event: i3ipc.events.WindowEvent):
    if event.change == "focus":
        on_window_focus(ipc, event)
    elif event.change == "close":
        on_window_close(ipc, event)


if __name__ == "__main__":
    default_layout: Optional[int] = None
    if len(sys.argv) == 2:
        if sys.argv[1].isnumeric():
            default_layout = int(sys.argv[1])
        else:
            print(f"Expected an integer, got: {sys.argv[1]}", file=sys.stderr)
            sys.exit(2)
    elif len(sys.argv) > 2:
        print("Too many arguments", file=sys.stderr)
        sys.exit(2)

    ipc = i3ipc.Connection()
    focused = ipc.get_tree().find_focused()
    if focused:
        prev_focused = focused.id
    else:
        prev_focused = None
    windows: dict = {}

    ipc.on("window", on_window)
    ipc.main()



#!/usr/bin/env python3

# This script requires i3ipc-python package (install it from a system package
# manager or pip).
# The script "stacks" numbered workspaces 1-10 onto 11-20, 21-30, etc
# Useful for hoarding workspaces, or if you're working on a project and need
# to focus on something else, but don't wanna close your workspaces
# This script doesn't proivde a way to view workspaces in the stack, nor does it
# expect you to provide one in your sway config. The stack is for storage; pop
# workspaces onto the "home row" (1-10) to view them.

# Example sway config:
#
# set $mod Mod4
# set $alt Mod1
# bindsym $mod+$alt+s exec /path/to/swaystack.py --up
# bindsym $mod+$alt+d exec /path/to/swaystack.py --down
# bindsym $mod+$alt+r exec /path/to/swaystack.py --rot-down
# bindsym $mod+$alt+e exec /path/to/swaystack.py --rot-up

import argparse

import i3ipc


def shift_up(workspace):
    workspace_num = workspace.num

    # only call from "home row"
    if not (workspace_num >= 1 and workspace_num <= 10):
        return

    stack_num = workspace_num % 10
    stack = (w for w in ipc.get_workspaces() 
             if w.num % 10 == stack_num and w.num != 0)
    sorted_stack = sorted(stack, key=lambda w: w.num, reverse=True)

    if workspace.leaves():
        # if not empty, push up into stack
        for ws in sorted_stack:
            num = ws.num
            ipc.command(f"rename workspace {num} to {num+10}")
        ipc.command(f"workspace {workspace_num}")
    # else, do nothing


def shift_down(workspace):
    workspace_num = workspace.num

    # only call from "home row"
    if not (workspace_num >= 1 and workspace_num <= 10):
        return

    stack_num = workspace_num % 10
    stack = (w for w in ipc.get_workspaces() 
             if w.num % 10 == stack_num and w.num > 10)
    sorted_stack = sorted(stack, key=lambda w: w.num)

    if not workspace.leaves():
        bottom_num = sorted_stack[0].num
        ipc.command(f"workspace {bottom_num}")
        # if empty, pop down from the stack
        for ws in sorted_stack:
            num = ws.num
            ipc.command(f"rename workspace {num} to {num-10}")
    # else, do nothing


def rotate_down(workspace):
    workspace_num = workspace.num

    # only call from "home row"
    if not (workspace_num >= 1 and workspace_num <= 10):
        return

    stack_num = workspace_num % 10

    # TODO: cleanup repeat code
    if workspace.leaves():
        # if workspace not empty, place on opposite end of stack
        stack = (w for w in ipc.get_workspaces() 
                 if w.num % 10 == stack_num and w.num != 0)
        top_stack = max(stack, key=lambda w: w.num)
        top_num = top_stack.num

        ipc.command(f"rename workspace {workspace_num} to {top_num+10}")

    stack = (w for w in ipc.get_workspaces() 
             if w.num % 10 == stack_num and w.num > 10)
    sorted_stack = sorted(stack, key=lambda w: w.num)

    bottom_num = sorted_stack[0].num
    ipc.command(f"workspace {bottom_num}")
    for ws in sorted_stack:
        num = ws.num
        ipc.command(f"rename workspace {num} to {num-10}")


def rotate_up(workspace):
    workspace_num = workspace.num

    # only call from "home row"
    if not (workspace_num >= 1 and workspace_num <= 10):
        return

    stack_num = workspace_num % 10

    # TODO: cleanup repeat code
    if workspace.leaves():
        stack = (w for w in ipc.get_workspaces() 
                 if w.num % 10 == stack_num and w.num != 0)
        sorted_stack = sorted(stack, key=lambda w: w.num, reverse=True)

        for ws in sorted_stack:
            num = ws.num
            ipc.command(f"rename workspace {num} to {num+10}")

    stack = (w for w in ipc.get_workspaces() 
             if w.num % 10 == stack_num and w.num > 10)
    top_stack = max(stack, key=lambda w: w.num)
    top_num = top_stack.num

    ipc.command(f"workspace {top_num}")
    ipc.command(f"rename workspace {top_num} to {workspace_num}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Workspace stacking, for hoarding workspaces. Requires "
        "numerical workspaces 1-10."
    )
    action = parser.add_mutually_exclusive_group()
    action.add_argument(
        "--up",
        action="store_true",
        help="Push non-empty focused workspace onto stack (default)",
    )
    action.add_argument(
        "--down",
        action="store_true",
        help="Pop top of stack onto empty focused workspace",
    )
    action.add_argument(
        "--rot-down",
        action="store_true",
        help="Rotate down along the focused workspace stack",
    )
    action.add_argument(
        "--rot-up",
        action="store_true",
        help="Rotate up along the focused workspace stack",
    )
    args = parser.parse_args()

    ipc = i3ipc.Connection()
    focused = ipc.get_tree().find_focused().workspace()

    if args.down:
        shift_down(focused)
    elif args.rot_down:
        rotate_down(focused)
    elif args.rot_up:
        rotate_up(focused)
    else:
        shift_up(focused)
